use super::common::flag;
use super::integer::{Time,Ratio64,Planar64,Planar64Vec3};

pub type Controls=u32;

#[binrw::binrw]
#[brw(little)]
pub struct StyleModifiers{
	pub header:u8,
	pub controls_mask:Controls,
	pub controls_mask_state:Controls,
	#[br(if(header&Self::STRAFE!=0))]
	pub strafe:Option<StrafeSettings>,
	#[br(if(header&Self::ROCKET!=0))]
	pub rocket:Option<PropulsionSettings>,
	#[br(if(header&Self::JUMP!=0))]
	pub jump:Option<JumpSettings>,
	#[br(if(header&Self::WALK!=0))]
	pub walk:Option<WalkSettings>,
	#[br(if(header&Self::LADDER!=0))]
	pub ladder:Option<LadderSettings>,
	#[br(if(header&Self::SWIM!=0))]
	pub swim:Option<PropulsionSettings>,
	pub gravity:Planar64Vec3,
	pub hitbox:Hitbox,
	pub camera_offset:Planar64Vec3,
	pub mass:Planar64,
}
impl StyleModifiers{
	const STRAFE:u8=1<<0;
	const ROCKET:u8=1<<1;
	const JUMP:u8=1<<2;
	const WALK:u8=1<<3;
	const LADDER:u8=1<<4;
	const SWIM:u8=1<<5;
}
impl Into<strafesnet_common::gameplay_style::StyleModifiers> for StyleModifiers{
	fn into(self)->strafesnet_common::gameplay_style::StyleModifiers{
		strafesnet_common::gameplay_style::StyleModifiers{
			//TODO: fail gracefully in binrw instead of panicing here
			controls_mask:strafesnet_common::controls_bitflag::Controls::from_bits(self.controls_mask).unwrap(),
			controls_mask_state:strafesnet_common::controls_bitflag::Controls::from_bits(self.controls_mask_state).unwrap(),
			strafe:self.strafe.map(Into::into),
			rocket:self.rocket.map(Into::into),
			jump:self.jump.map(Into::into),
			walk:self.walk.map(Into::into),
			ladder:self.ladder.map(Into::into),
			swim:self.swim.map(Into::into),
			gravity:strafesnet_common::integer::Planar64Vec3::raw_array(self.gravity),
			hitbox:self.hitbox.into(),
			camera_offset:strafesnet_common::integer::Planar64Vec3::raw_array(self.camera_offset),
			mass:strafesnet_common::integer::Planar64::raw(self.mass),
		}
	}
}
impl From<strafesnet_common::gameplay_style::StyleModifiers> for StyleModifiers{
	fn from(value:strafesnet_common::gameplay_style::StyleModifiers)->Self{
		let header=
			flag(value.strafe.is_some(),StyleModifiers::STRAFE)
			|flag(value.rocket.is_some(),StyleModifiers::ROCKET)
			|flag(value.jump.is_some(),StyleModifiers::JUMP)
			|flag(value.walk.is_some(),StyleModifiers::WALK)
			|flag(value.ladder.is_some(),StyleModifiers::LADDER)
			|flag(value.swim.is_some(),StyleModifiers::SWIM);
		Self{
			header,
			controls_mask:value.controls_mask.bits(),
			controls_mask_state:value.controls_mask_state.bits(),
			strafe:value.strafe.map(Into::into),
			rocket:value.rocket.map(Into::into),
			jump:value.jump.map(Into::into),
			walk:value.walk.map(Into::into),
			ladder:value.ladder.map(Into::into),
			swim:value.swim.map(Into::into),
			gravity:value.gravity.get().to_array(),
			hitbox:value.hitbox.into(),
			camera_offset:value.camera_offset.get().to_array(),
			mass:value.mass.get(),
		}
	}
}

#[binrw::binrw]
#[brw(little,repr=u8)]
pub enum JumpCalculation{
	Capped,
	Energy,
	Linear,
}
impl Into<strafesnet_common::gameplay_style::JumpCalculation> for JumpCalculation{
	fn into(self)->strafesnet_common::gameplay_style::JumpCalculation{
		match self{
			JumpCalculation::Capped=>strafesnet_common::gameplay_style::JumpCalculation::Capped,
			JumpCalculation::Energy=>strafesnet_common::gameplay_style::JumpCalculation::Energy,
			JumpCalculation::Linear=>strafesnet_common::gameplay_style::JumpCalculation::Linear,
		}
	}
}
impl From<strafesnet_common::gameplay_style::JumpCalculation> for JumpCalculation{
	fn from(value:strafesnet_common::gameplay_style::JumpCalculation)->Self{
		match value{
			strafesnet_common::gameplay_style::JumpCalculation::Capped=>JumpCalculation::Capped,
			strafesnet_common::gameplay_style::JumpCalculation::Energy=>JumpCalculation::Energy,
			strafesnet_common::gameplay_style::JumpCalculation::Linear=>JumpCalculation::Linear,
		}
	}
}

pub enum JumpImpulse{
	FromTime(Time),
	FromHeight(Planar64),
	FromDeltaV(Planar64),
	FromEnergy(Planar64),
}
impl Into<strafesnet_common::gameplay_style::JumpImpulse> for JumpImpulse{
	fn into(self)->strafesnet_common::gameplay_style::JumpImpulse{
		match self{
			JumpImpulse::FromTime(time)=>strafesnet_common::gameplay_style::JumpImpulse::FromTime(strafesnet_common::integer::Time::raw(time)),
			JumpImpulse::FromHeight(height)=>strafesnet_common::gameplay_style::JumpImpulse::FromHeight(strafesnet_common::integer::Planar64::raw(height)),
			JumpImpulse::FromDeltaV(deltav)=>strafesnet_common::gameplay_style::JumpImpulse::FromDeltaV(strafesnet_common::integer::Planar64::raw(deltav)),
			JumpImpulse::FromEnergy(energy)=>strafesnet_common::gameplay_style::JumpImpulse::FromEnergy(strafesnet_common::integer::Planar64::raw(energy)),
		}
	}
}
impl From<strafesnet_common::gameplay_style::JumpImpulse> for JumpImpulse{
	fn from(value:strafesnet_common::gameplay_style::JumpImpulse)->Self{
		match value{
			strafesnet_common::gameplay_style::JumpImpulse::FromTime(time)=>JumpImpulse::FromTime(time.get()),
			strafesnet_common::gameplay_style::JumpImpulse::FromHeight(height)=>JumpImpulse::FromHeight(height.get()),
			strafesnet_common::gameplay_style::JumpImpulse::FromDeltaV(deltav)=>JumpImpulse::FromDeltaV(deltav.get()),
			strafesnet_common::gameplay_style::JumpImpulse::FromEnergy(energy)=>JumpImpulse::FromEnergy(energy.get()),
		}
	}
}

#[binrw::binrw]
#[brw(little)]
pub struct ControlsActivation{
	controls_mask:Controls,
	controls_intersects:Controls,
	controls_contains:Controls,
}
impl Into<strafesnet_common::gameplay_style::ControlsActivation> for ControlsActivation{
	fn into(self)->strafesnet_common::gameplay_style::ControlsActivation{
		strafesnet_common::gameplay_style::ControlsActivation::new(
			strafesnet_common::controls_bitflag::Controls::from_bits(self.controls_mask).unwrap(),
			strafesnet_common::controls_bitflag::Controls::from_bits(self.controls_intersects).unwrap(),
			strafesnet_common::controls_bitflag::Controls::from_bits(self.controls_contains).unwrap(),
		)
	}
}
impl From<strafesnet_common::gameplay_style::ControlsActivation> for ControlsActivation{
	fn from(value:strafesnet_common::gameplay_style::ControlsActivation)->Self{
		Self{
			controls_mask:value.controls_mask().bits(),
			controls_intersects:value.controls_intersects().bits(),
			controls_contains:value.controls_contains().bits(),
		}
	}
}

#[binrw::binrw]
#[brw(little)]
pub struct StrafeSettings{
	header:u8,
	enable:ControlsActivation,
	mv:Planar64,
	#[br(if(header&Self::AIR_ACCEL_LIMIT!=0))]
	air_accel_limit:Option<Planar64>,
	tick_rate:Ratio64,
}
impl StrafeSettings{
	const AIR_ACCEL_LIMIT:u8=1<<0;
}
impl Into<strafesnet_common::gameplay_style::StrafeSettings> for StrafeSettings{
	fn into(self)->strafesnet_common::gameplay_style::StrafeSettings{
		strafesnet_common::gameplay_style::StrafeSettings::new(
			self.enable.into(),
			strafesnet_common::integer::Planar64::raw(self.mv),
			self.air_accel_limit.map(strafesnet_common::integer::Planar64::raw),
			self.tick_rate.into(),
		)
	}
}
impl From<strafesnet_common::gameplay_style::StrafeSettings> for StrafeSettings{
	fn from(value:strafesnet_common::gameplay_style::StrafeSettings)->Self{
		let (enable,mv,air_accel_limit,tick_rate)=value.into_inner();
		let header=flag(air_accel_limit.is_some(),StrafeSettings::AIR_ACCEL_LIMIT);
		Self{
			header,
			enable:enable.into(),
			mv:mv.get(),
			air_accel_limit:air_accel_limit.map(|a|a.get()),
			tick_rate:tick_rate.into(),
		}
	}
}

#[binrw::binrw]
#[brw(little)]
pub struct PropulsionSettings{
	magnitude:Planar64,
}
impl Into<strafesnet_common::gameplay_style::PropulsionSettings> for PropulsionSettings{
	fn into(self)->strafesnet_common::gameplay_style::PropulsionSettings{
		strafesnet_common::gameplay_style::PropulsionSettings::new(
			strafesnet_common::integer::Planar64::raw(self.magnitude)
		)
	}
}
impl From<strafesnet_common::gameplay_style::PropulsionSettings> for PropulsionSettings{
	fn from(value:strafesnet_common::gameplay_style::PropulsionSettings)->Self{
		Self{
			magnitude:value.magnitude().get(),
		}
	}
}

#[binrw::binrw]
#[brw(little)]
pub struct JumpSettings{
	header:u8,
	impulse:i64,
}
impl JumpSettings{
	const IMPULSE:u8=0b0011;
	const CALCULATION:u8=0b1100;
	const fn impulse(&self)->Option<strafesnet_common::gameplay_style::JumpImpulse>{
		match self.header&Self::IMPULSE{
			0=>Some(strafesnet_common::gameplay_style::JumpImpulse::FromTime(strafesnet_common::integer::Time::raw(self.impulse))),
			1=>Some(strafesnet_common::gameplay_style::JumpImpulse::FromHeight(strafesnet_common::integer::Planar64::raw(self.impulse))),
			2=>Some(strafesnet_common::gameplay_style::JumpImpulse::FromDeltaV(strafesnet_common::integer::Planar64::raw(self.impulse))),
			3=>Some(strafesnet_common::gameplay_style::JumpImpulse::FromEnergy(strafesnet_common::integer::Planar64::raw(self.impulse))),
			_=>None,
		}
	}
	const fn calculation(&self)->Option<strafesnet_common::gameplay_style::JumpCalculation>{
		match (self.header&Self::CALCULATION)>>2{
			0=>Some(strafesnet_common::gameplay_style::JumpCalculation::Capped),
			1=>Some(strafesnet_common::gameplay_style::JumpCalculation::Energy),
			2=>Some(strafesnet_common::gameplay_style::JumpCalculation::Linear),
			_=>None,
		}
	}
}
impl Into<strafesnet_common::gameplay_style::JumpSettings> for JumpSettings{
	fn into(self)->strafesnet_common::gameplay_style::JumpSettings{
		strafesnet_common::gameplay_style::JumpSettings::new(
			self.impulse().unwrap(),
			self.calculation().unwrap(),
		)
	}
}
impl From<strafesnet_common::gameplay_style::JumpSettings> for JumpSettings{
	fn from(value:strafesnet_common::gameplay_style::JumpSettings)->Self{
		let (impulse_enum,calculation)=value.into_inner();
		let (impulse,impulse_header)=match impulse_enum{
			strafesnet_common::gameplay_style::JumpImpulse::FromTime(impulse)=>(impulse.get(),0),
			strafesnet_common::gameplay_style::JumpImpulse::FromHeight(impulse)=>(impulse.get(),1),
			strafesnet_common::gameplay_style::JumpImpulse::FromDeltaV(impulse)=>(impulse.get(),2),
			strafesnet_common::gameplay_style::JumpImpulse::FromEnergy(impulse)=>(impulse.get(),3),
		};
		let calculation_header=match calculation{
			strafesnet_common::gameplay_style::JumpCalculation::Capped=>0,
			strafesnet_common::gameplay_style::JumpCalculation::Energy=>1,
			strafesnet_common::gameplay_style::JumpCalculation::Linear=>2,
		};
		let header=
			impulse_header
			|(calculation_header<<2);
		Self{
			header,
			impulse,
		}
	}
}

#[binrw::binrw]
#[brw(little)]
pub struct AccelerateSettings{
	accel:Planar64,
	topspeed:Planar64,
}
impl Into<strafesnet_common::gameplay_style::AccelerateSettings> for AccelerateSettings{
	fn into(self)->strafesnet_common::gameplay_style::AccelerateSettings{
		strafesnet_common::gameplay_style::AccelerateSettings::new(
			strafesnet_common::integer::Planar64::raw(self.accel),
			strafesnet_common::integer::Planar64::raw(self.topspeed),
		)
	}
}
impl From<strafesnet_common::gameplay_style::AccelerateSettings> for AccelerateSettings{
	fn from(value:strafesnet_common::gameplay_style::AccelerateSettings)->Self{
		Self{
			accel:value.accel().get(),
			topspeed:value.topspeed().get(),
		}
	}
}

#[binrw::binrw]
#[brw(little)]
pub struct WalkSettings{
	accelerate:AccelerateSettings,
	static_friction:Planar64,
	kinetic_friction:Planar64,
	surf_dot:Planar64,
}
impl Into<strafesnet_common::gameplay_style::WalkSettings> for WalkSettings{
	fn into(self)->strafesnet_common::gameplay_style::WalkSettings{
		strafesnet_common::gameplay_style::WalkSettings::new(
			self.accelerate.into(),
			strafesnet_common::integer::Planar64::raw(self.static_friction),
			strafesnet_common::integer::Planar64::raw(self.kinetic_friction),
			strafesnet_common::integer::Planar64::raw(self.surf_dot),
		)
	}
}
impl From<strafesnet_common::gameplay_style::WalkSettings> for WalkSettings{
	fn from(value:strafesnet_common::gameplay_style::WalkSettings)->Self{
		let (accelerate,static_friction,kinetic_friction,surf_dot)=value.into_inner();
		Self{
			accelerate:accelerate.into(),
			static_friction:static_friction.get(),
			kinetic_friction:kinetic_friction.get(),
			surf_dot:surf_dot.get(),
		}
	}
}

#[binrw::binrw]
#[brw(little)]
pub struct LadderSettings{
	accelerate:AccelerateSettings,
	dot:Planar64,
}
impl Into<strafesnet_common::gameplay_style::LadderSettings> for LadderSettings{
	fn into(self)->strafesnet_common::gameplay_style::LadderSettings{
		strafesnet_common::gameplay_style::LadderSettings::new(
			self.accelerate.into(),
			strafesnet_common::integer::Planar64::raw(self.dot),
		)
	}
}
impl From<strafesnet_common::gameplay_style::LadderSettings> for LadderSettings{
	fn from(value:strafesnet_common::gameplay_style::LadderSettings)->Self{
		let (accelerate,dot)=value.into_inner();
		Self{
			accelerate:accelerate.into(),
			dot:dot.get(),
		}
	}
}

#[binrw::binrw]
#[brw(little,repr=u8)]
pub enum HitboxMesh{
	Box,
	Cylinder,
}
impl Into<strafesnet_common::gameplay_style::HitboxMesh> for HitboxMesh{
	fn into(self)->strafesnet_common::gameplay_style::HitboxMesh{
		match self{
			HitboxMesh::Box=>strafesnet_common::gameplay_style::HitboxMesh::Box,
			HitboxMesh::Cylinder=>strafesnet_common::gameplay_style::HitboxMesh::Cylinder,
		}
	}
}
impl From<strafesnet_common::gameplay_style::HitboxMesh> for HitboxMesh{
	fn from(value:strafesnet_common::gameplay_style::HitboxMesh)->Self{
		match value{
			strafesnet_common::gameplay_style::HitboxMesh::Box=>HitboxMesh::Box,
			strafesnet_common::gameplay_style::HitboxMesh::Cylinder=>HitboxMesh::Cylinder,
		}
	}
}

#[binrw::binrw]
#[brw(little)]
pub struct Hitbox{
	pub halfsize:Planar64Vec3,
	pub mesh:HitboxMesh,
}
impl Into<strafesnet_common::gameplay_style::Hitbox> for Hitbox{
	fn into(self)->strafesnet_common::gameplay_style::Hitbox{
		strafesnet_common::gameplay_style::Hitbox{
			halfsize:strafesnet_common::integer::Planar64Vec3::raw_array(self.halfsize),
			mesh:self.mesh.into(),
		}
	}
}
impl From<strafesnet_common::gameplay_style::Hitbox> for Hitbox{
	fn from(value:strafesnet_common::gameplay_style::Hitbox)->Self{
		Self{
			halfsize:value.halfsize.get().to_array(),
			mesh:value.mesh.into(),
		}
	}
}
